/*
Copyright (C) 2003 David Bateman

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

In addition to the terms of the GPL, you are permitted to link
this program with any Open Source program, as defined by the
Open Source Initiative (www.opensource.org)
*/

#include <octave/oct.h>
#include <octave/lo-mappers.h>
#include <octave/config.h>
#include <octave/quit.h>
#include <octave/lo-ieee.h>
#include <octave/data-conv.h>
#include "oct-sort.cc"

template <class T>
class
vec_index
{
 public:
  vec_index (void) : vec (0), indx (-1) { }
  vec_index (T v, int i) : vec (v), indx (i) { }
  vec_index<T> &operator = (const vec_index<T> &x);
  T vec;
  int indx;
};

template <class T>
vec_index<T> &vec_index<T>::operator = (const vec_index<T> &x)
{
  vec = x.vec;
  indx = x.indx;
  return (*this);
}

#ifdef IEEE754

// If we don't have eight byte integers defined we can't use this code
#ifndef EIGHT_BYTE_INT
#error ("No 8-byte integer type available")
#endif

static inline unsigned EIGHT_BYTE_INT FloatFlip(unsigned EIGHT_BYTE_INT f)
{
  unsigned EIGHT_BYTE_INT mask = -(EIGHT_BYTE_INT)(f >> 63) | 
    0x8000000000000000ULL;
  return f ^ mask;
}

inline unsigned EIGHT_BYTE_INT IFloatFlip(unsigned EIGHT_BYTE_INT f)
{
  unsigned EIGHT_BYTE_INT mask = ((f >> 63) - 1) | 0x8000000000000000ULL;
  return f ^ mask;
}

template vec_index<unsigned EIGHT_BYTE_INT>;

bool
operator < (const vec_index<unsigned EIGHT_BYTE_INT> &a, 
	    const vec_index<unsigned EIGHT_BYTE_INT> &b)
{
  return (a.vec < b.vec);
}

template octave_sort<unsigned EIGHT_BYTE_INT>;
template octave_sort<vec_index<unsigned EIGHT_BYTE_INT> >;
#else
template vec_index<double>;

bool
operator < (const vec_index<double> &a, const vec_index<double> &b)
{
  return (xisnan(b.vec) || (a.vec < b.vec));
}

bool
double_compare (double a, double b)
{
  return (xisnan(b) || (a < b));
}

template octave_sort<double>;
template octave_sort< vec_index<double> >;
#endif

static octave_value_list
vec_sort (RowVector &vec, bool return_idx)
{

  octave_value_list retval;
  int elements = vec.capacity ();
  if (elements == 1)
    {
      if (return_idx)
	retval(1) = RowVector (elements, 1.0);
      retval(0) = vec;
      return retval;
    }
  else if (elements > 1)
    {
#ifdef IEEE754
      /* Flip the data in the vector so that int compares on IEE754 give
       * the correct ordering
       */
      unsigned EIGHT_BYTE_INT *p = 
	(unsigned EIGHT_BYTE_INT *)vec.fortran_vec ();
      for (int i = 0; i < elements; i++)
	p[i] = FloatFlip (p[i]);

      if (return_idx)
	{
	  // Is there a better way of writing indexed sorts without
	  //  exploding the octave_sort class into lots of special cases?

	  octave_sort<vec_index<unsigned EIGHT_BYTE_INT> > 
	    indexed_ieee754_sort;

	  OCTAVE_LOCAL_BUFFER (vec_index<unsigned EIGHT_BYTE_INT>, 
			       vi, elements);

	  for (int i = 0; i < elements; i++)
	    {
	      vi[i].vec = p[i];
	      vi[i].indx = i + 1;
	    }

	  indexed_ieee754_sort.sort (vi, elements);

	  RowVector idx (elements);

	  for (int i = 0; i < elements; i++)
	    {
	      p[i] = vi[i].vec;
	      idx(i) = vi[i].indx;
	    }

	  /* Flip the data out of the vector so that int compares on
	   *  IEE754 give the correct ordering
	   */
	  for (int i = 0; i < elements; i++)
	    p[i] = IFloatFlip (p[i]);

	  /* There are two representations of NaN. One will be sorted to the
	   * beginning of the vector and the other to the end. If it will be
	   * sort to the beginning, fix things up.
	   */
	  if (lo_ieee_signbit (octave_NaN))
	    {
	      int k;
	      unsigned int i = 0;
	      double *v = vec.fortran_vec ();
	      while (xisnan(v[i++]));
	      OCTAVE_LOCAL_BUFFER (double, itmp, i - 1);
	      for (unsigned int j = 0; j < i -1; j++)
		itmp[i] = idx(j);
	      for (unsigned int j = 0; j < elements - i + 1; j++)
		{
		  v[j] = v[j+i-1];
		  idx(j) = idx(j+i-1);
		}
	      for (unsigned int j = elements - i + 1, k = 0; j < elements; 
		   j++, k++)
		{
		  v[j] = octave_NaN;
		  idx(j) = itmp[k];
		}
	    }
	  retval (1) = idx;
	}
      else
	{
	  octave_sort<unsigned EIGHT_BYTE_INT> ieee754_sort;
 
	  ieee754_sort.sort (p, elements);

	  /* Flip the data out of the vector so that int compares on
	   *  IEE754 give the correct ordering
	   */
	  for (int i = 0; i < elements; i++)
	    p[i] = IFloatFlip (p[i]);

	  /* There are two representations of NaN. One will be sorted to the
	   * beginning of the vector and the other to the end. If it will be
	   * sort to the beginning, fix things up.
	   */
	  if (lo_ieee_signbit (octave_NaN))
	    {
	      unsigned int i = 0;
	      double *v = vec.fortran_vec ();
	      while (xisnan(v[i++]));
	      for (unsigned int j = 0; j < elements - i + 1; j++)
		v[j] = v[j+i-1];
	      for (unsigned int j = elements - i + 1; j < elements; j++)
		v[j] = octave_NaN;
	    }
	}
#else
      if (return_idx)
	{
	  // Is there a better way of writing indexed sorts without
	  //  exploding the octave_sort class into lots of special cases?

	  octave_sort<vec_index<double> > indexed_double_sort;

	  OCTAVE_LOCAL_BUFFER (vec_index<double>, vi, elements);

	  for (int i = 0; i < elements; i++)
	    {
	      vi[i].vec = vec(i);
	      vi[i].indx = i + 1;
	    }

	  indexed_double_sort.sort (vi, elements);

	  RowVector idx (elements);

	  for (int i = 0; i < elements; i++)
	    {
	      vec(i) = vi[i].vec;
	      idx(i) = vi[i].indx;
	    }

	  retval (1) = idx;

	}
      else
	{
	  double *v = vec.fortran_vec ();
	  octave_sort<double> double_sort (double_compare);
	  double_sort.sort (v, elements);
	}
#endif

      retval(0) = vec;
    }

  return retval;
}

static octave_value_list
mx_sort (Matrix &m, bool return_idx)
{
  octave_value_list retval;
  int nr = m.rows ();
  int nc = m.columns ();

  if (nr == 1 && nc > 0)
    {
      if (return_idx)
	retval(1) = Matrix (nr, nc, 1.0);
      retval(0) = m;
      return retval;
    }
  else if (nr > 1 && nc > 0)
    {

#ifdef IEEE754
      double *v = m.fortran_vec ();

      unsigned EIGHT_BYTE_INT *p = (unsigned EIGHT_BYTE_INT *)v;

      if (return_idx)
	{
	  // Is there a better way of writing indexed sorts without
	  // exploding the octave_sort class into lots of special cases?

	  octave_sort<vec_index<unsigned EIGHT_BYTE_INT> > 
	    indexed_ieee754_sort;

	  OCTAVE_LOCAL_BUFFER (vec_index<unsigned EIGHT_BYTE_INT>, 
			       vi, nr);

	  Matrix idx (nr, nc);

	  for (int j = 0; j < nc; j++)
	    {

	      /* Flip the data in the vector so that int compares on 
	       * IEE754 give the correct ordering
	       */
	      for (int i = 0; i < nr; i++)
		{
		  vi[i].vec = FloatFlip (p[i]);
		  vi[i].indx = i + 1;
		}

	      indexed_ieee754_sort.sort (vi, nr);

	      /* Flip the data out of the vector so that int compares on
	       *  IEE754 give the correct ordering
	       */
	      for (int i = 0; i < nr; i++)
		{
		  p[i] = IFloatFlip (vi[i].vec);
		  idx(i,j) = vi[i].indx;
		}

	      /* There are two representations of NaN. One will be sorted to
	       * the beginning of the vector and the other to the end. If it
	       * will be sort to the beginning, fix things up.
	       */
	      if (lo_ieee_signbit (octave_NaN))
		{
		  unsigned int i = 0;
		  double *vtmp = (double *)p;
		  while (xisnan(vtmp[i++]));
		  OCTAVE_LOCAL_BUFFER (double, itmp, i - 1);
		  for (unsigned int l = 0; l < i -1; l++)
		    itmp[l] = idx(l,j);
		  for (unsigned int l = 0; l < nr - i + 1; l++)
		    {
		      vtmp[l] = vtmp[l+i-1];
		      idx(l,j) = idx(l+i-1,j);
		    }
		  for (unsigned int k = 0, l = nr - i + 1; l < nr; 
		       l++, k++)
		    {
		      vtmp[l] = octave_NaN;
		      idx(l,j) = itmp[k];
		    }
		}

	      p += nr;
	    }

	  retval (1) = idx;
	}
      else
	{
	  octave_sort<unsigned EIGHT_BYTE_INT> ieee754_sort;
 
	  for (int j = 0; j < nc; j++)
	    {
	      /* Flip the data in the vector so that int compares on 
	       * IEE754 give the correct ordering
	       */
	      for (int i = 0; i < nr; i++)
		p[i] = FloatFlip (p[i]);

	      ieee754_sort.sort (p, nr);

	      /* Flip the data out of the vector so that int compares on
	       *  IEE754 give the correct ordering
	       */
	      for (int i = 0; i < nr; i++)
		p[i] = IFloatFlip (p[i]);

	      /* There are two representations of NaN. One will be sorted to
	       * the beginning of the vector and the other to the end. If it
	       * will be sort to the beginning, fix things up.
	       */
	      if (lo_ieee_signbit (octave_NaN))
		{
		  unsigned int i = 0;
		  double *vtmp = (double *)p;
		  while (xisnan(vtmp[i++]));
		  for (unsigned int j = 0; j < nr - i + 1; j++)
		    vtmp[j] = vtmp[j+i-1];
		  for (unsigned int j = nr - i + 1; j < nr; j++)
		    vtmp[j] = octave_NaN;
		}

	      p += nr;
	    }
	}
#else
      if (return_idx)
	{
	  // Is there a better way of writing indexed sorts without
	  //  exploding the octave_sort class into lots of special cases?

	  octave_sort<vec_index<double> > indexed_double_sort;

	  OCTAVE_LOCAL_BUFFER (vec_index<double>, vi, nr);

	  Matrix idx (nr, nc);

	  for (int j = 0; j < nc; j++)
	    {
	      for (int i = 0; i < nr; i++)
		{
		  vi[i].vec = m(i,j);
		  vi[i].indx = i + 1;
		}

	      indexed_double_sort.sort (vi, nr);

	      for (int i = 0; i < nr; i++)
		{
		  m(i,j) = vi[i].vec;
		  idx(i,j) = vi[i].indx;
		}
	    }
	  retval (1) = idx;
	}
      else
	{
	  double *v = m.fortran_vec ();
	  octave_sort<double> double_sort (double_compare);
	  for (int j = 0; j < nc; j++) 
	    {
	      double_sort.sort (v, nr);
	      v += nr;
	    }
	}
#endif
      retval(0) = m;
    }

  return retval;
}

#ifdef IEEE754
DEFUN_DLD (merge_ieee754, args, nargout,
#else
DEFUN_DLD (merge_double, args, nargout,
#endif
  "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {[@var{s}, @var{i}] =} sort (@var{x})\n\
Return a copy of @var{x} with the elements elements arranged in\n\
increasing order.  For matrices, @code{sort} orders the elements in each\n\
column.\n\
\n\
For example,\n\
\n\
@example\n\
@group\n\
sort ([1, 2; 2, 3; 3, 1])\n\
     @result{}  1  1\n\
         2  2\n\
         3  3\n\
@end group\n\
@end example\n\
\n\
The @code{sort} function may also be used to produce a matrix\n\
containing the original row indices of the elements in the sorted\n\
matrix.  For example,\n\
\n\
@example\n\
@group\n\
[s, i] = sort ([1, 2; 2, 3; 3, 1])\n\
     @result{} s = 1  1\n\
            2  2\n\
            3  3\n\
     @result{} i = 1  3\n\
            2  1\n\
            3  2\n\
@end group\n\
@end example\n\
\n\
The for equal elements, the indices are such that the equal elements\n\
are listed in the order that appeared in the original list.\n\
@end deftypefn")
{
  octave_value_list retval;

  int nargin = args.length ();

  if (nargin != 1)
    {
      print_usage ("sort");
      return retval;
    }

  bool return_idx = nargout > 1;

  octave_value arg = args(0);

  if (arg.is_real_type ())
    {
      Matrix m = arg.matrix_value ();

      if (! error_state)
	{
	  if (m.rows () == 1)
	    {
	      int nc = m.columns ();
	      RowVector v (nc);
	      for (int i = 0; i < nc; i++)
		v(i) = m(0,i);

	      retval = vec_sort (v, return_idx);
	    }
	  else
	    retval = mx_sort (m, return_idx);
	}
    }
  else if (arg.is_complex_type ())
    error ("merge: can not sort complex type yet");
  else
    gripe_wrong_type_arg ("sort", arg);

  return retval;
}

/*
;;; Local Variables: ***
;;; mode: C++ ***
;;; End: ***
*/
