/*
Copyright (C) 2003 David Bateman

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

In addition to the terms of the GPL, you are permitted to link
this program with any Open Source program, as defined by the
Open Source Initiative (www.opensource.org)

This code is based liberally on the radix.cpp and bore the following 
copyright

 Radix.cpp: a fast floating-point radix sort demo

   Copyright (C) Herf Consulting LLC 2001.  All Rights Reserved.
   Use for anything you want, just tell me what you do with it.
   Code provided "as-is" with no liabilities for anything that goes wrong.
*/

#include <octave/oct.h>
#include <octave/lo-mappers.h>
#include <octave/config.h>
#include <octave/quit.h>
#include <octave/lo-ieee.h>
#include <octave/data-conv.h>

// If we don't have eight byte integers defined we can't use this code
#ifndef EIGHT_BYTE_INT
#error ("No 8-byte integer type available")
#endif

static Array<int>
create_index_array (int n)
{
  Array<int> l (n);
  for (int i = 0; i < n; i++)
    l (i) = i+1;
  return l;
}

// Make my life easier
typedef EIGHT_BYTE_INT int64;
typedef unsigned EIGHT_BYTE_INT uint64;

// =======================================================================
// flip a float for sorting
//  finds SIGN of fp number.
//  if it's 1 (negative float), it flips all bits
//  if it's 0 (positive float), it flips the sign only
// =======================================================================
static inline uint64 FloatFlip(uint64 f)
{
  uint64 mask = -int64(f >> 63) | 0x8000000000000000ULL;
  return f ^ mask;
}

static inline void FloatFlipX(uint64 &f)
{
  uint64 mask = -int64(f >> 63) | 0x8000000000000000ULL;
  f ^= mask;
}

// =======================================================================
// flip a float back (invert FloatFlip)
//  signed was flipped from above, so:
//  if sign is 1 (negative), it flips the sign bit back
//  if sign is 0 (positive), it flips all bits back
// =======================================================================
inline uint64 IFloatFlip(uint64 f)
{
  uint64 mask = ((f >> 63) - 1) | 0x8000000000000000ULL;
  return f ^ mask;
}

// Initialize a memory space for the sort histograms. 2k elements or
// 8kB should be smaller than most L1 cache sizes, which is good
unsigned int histogram[2048];

// ---- utils for accessing 9-bit quantities in doubles
// This means that the 8 histograms are 256 elements in size, giving
// a total memory requirement of 8k bytes. This should be small
// enough to fit into the L1 cache on most processors..
#define _0(x)	(unsigned int)(x & 0xFF)
#define _1(x)	(unsigned int)(x >> 8 & 0xFF)
#define _2(x)	(unsigned int)(x >> 16 & 0xFF)
#define _3(x)	(unsigned int)(x >> 24 & 0xFF)
#define _4(x)	(unsigned int)(x >> 32 & 0xFF)
#define _5(x)	(unsigned int)(x >> 40 & 0xFF)
#define _6(x)	(unsigned int)(x >> 48 & 0xFF)
#define _7(x)	(unsigned int)(x >> 56)

// Define the macro DO_REAL_SORT for the core of the radix sort as
// it is used twice, once with and once without indexing
#define DO_REAL_SORT(elements, array, tmp)			\
{								\
  unsigned int i;						\
								\
  /* 8 histograms on the stack: */				\
  const unsigned int kHist = 256;				\
  unsigned int *b0 = histogram;					\
								\
  unsigned int *b1 = b0 + kHist;				\
  unsigned int *b2 = b1 + kHist;				\
  unsigned int *b3 = b2 + kHist;				\
  unsigned int *b4 = b3 + kHist;				\
  unsigned int *b5 = b4 + kHist;				\
  unsigned int *b6 = b5 + kHist;				\
  unsigned int *b7 = b6 + kHist;				\
								\
  for (i = 0; i < kHist * 8; i++)				\
    b0[i] = 0;							\
								\
  /* 1.  parallel histogramming pass */				\
  for (i = 0; i < elements; i++) 				\
    {								\
      uint64 fi = FloatFlip((uint64&)array[i]);			\
								\
      b0[_0(fi)] ++;						\
      b1[_1(fi)] ++;						\
      b2[_2(fi)] ++;						\
      b3[_3(fi)] ++;						\
      b4[_4(fi)] ++;						\
      b5[_5(fi)] ++;						\
      b6[_6(fi)] ++;						\
      b7[_7(fi)] ++;						\
    }								\
								\
  /* 2.  Sum the histograms -- each histogram entry records */	\
  /*     the number of values preceding itself.	            */	\
  {								\
    unsigned int sum0 = 0, sum1 = 0, sum2 = 0, sum3 = 0;	\
    unsigned int sum4 = 0, sum5 = 0, sum6 = 0, sum7 = 0;	\
    unsigned int tsum;						\
    for (i = 0; i < kHist; i++) 				\
      {								\
	tsum = b0[i] + sum0;					\
	b0[i] = sum0 - 1;					\
	sum0 = tsum;						\
								\
	tsum = b1[i] + sum1;					\
	b1[i] = sum1 - 1;					\
	sum1 = tsum;						\
								\
	tsum = b2[i] + sum2;					\
	b2[i] = sum2 - 1;					\
	sum2 = tsum;						\
								\
	tsum = b3[i] + sum3;					\
	b3[i] = sum3 - 1;					\
	sum3 = tsum;						\
								\
	tsum = b4[i] + sum4;					\
	b4[i] = sum4 - 1;					\
	sum4 = tsum;						\
								\
	tsum = b5[i] + sum5;					\
	b5[i] = sum5 - 1;					\
	sum5 = tsum;						\
								\
	tsum = b6[i] + sum6;					\
	b6[i] = sum6 - 1;					\
	sum6 = tsum;						\
								\
	tsum = b7[i] + sum7;					\
	b7[i] = sum7 - 1;					\
	sum7 = tsum;						\
      }								\
  }								\
								\
  /* byte 0: floatflip entire value, read/write histogram, */	\
  /* write out flipped                                     */	\
  for (i = 0; i < elements; i++) 				\
    {								\
      uint64 fi = array[i];					\
      FloatFlipX(fi);						\
      unsigned int pos = _0(fi);				\
      tmp[++b0[pos]] = fi;					\
      ARRAY_TO_TMP_EXTRA(b0[pos]);				\
    }								\
								\
  /* byte 1: read/write histogram, copy	*/			\
  /*   tmp -> array                     */			\
  for (i = 0; i < elements; i++) 				\
    {								\
      uint64 si = tmp[i];					\
      unsigned int pos = _1(si);				\
      array[++b1[pos]] = si;					\
      TMP_TO_ARRAY_EXTRA(b1[pos]);				\
    }								\
								\
  /* byte 2: read/write histogram, copy */			\
  /*   array -> tmp                     */			\
  for (i = 0; i < elements; i++) 				\
    {	 							\
      uint64 si = array[i]; 					\
      unsigned int pos = _2(si); 				\
      tmp[++b2[pos]] = si; 					\
      ARRAY_TO_TMP_EXTRA(b2[pos]);				\
    } 								\
 								\
  /* byte 3: read/write histogram, copy */			\
  /*   tmp -> array                     */			\
  for (i = 0; i < elements; i++)  				\
    { 								\
      uint64 si = tmp[i]; 					\
      unsigned int pos = _3(si); 				\
      array[++b3[pos]] = si; 					\
      TMP_TO_ARRAY_EXTRA(b3[pos]);				\
    } 								\
 								\
  /* byte 4: read/write histogram, copy */ 			\
  /*   array -> tmp                     */ 			\
  for (i = 0; i < elements; i++) 				\
   {  								\
      uint64 si = array[i]; 					\
      unsigned int pos = _4(si); 				\
      tmp[++b4[pos]] = si; 					\
      ARRAY_TO_TMP_EXTRA(b4[pos]);				\
    } 								\
 								\
  /* byte 5: read/write histogram, copy */ 			\
  /*   tmp -> array                     */ 			\
  for (i = 0; i < elements; i++) 				\
   {  								\
      uint64 si = tmp[i]; 					\
      unsigned int pos = _5(si); 				\
      array[++b5[pos]] = si; 					\
      TMP_TO_ARRAY_EXTRA(b5[pos]);				\
    } 								\
 								\
  /* byte 6: read/write histogram, copy */ 			\
  /*   array -> tmp                     */ 			\
  for (i = 0; i < elements; i++) 				\
   {  								\
      uint64 si = array[i]; 					\
      unsigned int pos = _6(si); 				\
      tmp[++b6[pos]] = si; 					\
      ARRAY_TO_TMP_EXTRA(b6[pos]);				\
    } 								\
 								\
  /* byte 7: read/write histogram, copy & flip out */ 		\
  /*   tmp -> array                                */		\
  for (i = 0; i < elements; i++)  				\
    {								\
      uint64 ai = tmp[i]; 					\
      unsigned int pos = _7(ai); 				\
      array[++b7[pos]] = IFloatFlip(ai);			\
      TMP_TO_ARRAY_EXTRA(b7[pos]);				\
    } 								\
}

static void
sort(double *x, double *tmp, int elements)
{
  uint64 *sort = (uint64*)tmp;
  uint64 *array = (uint64*)x;

#define ARRAY_TO_TMP_EXTRA(pos)
#define TMP_TO_ARRAY_EXTRA(pos)

  DO_REAL_SORT (elements, array, sort);

#undef ARRAY_TO_TMP_EXTRA
#undef TMP_TO_ARRAY_EXTRA

  // There are two representations of NaN. One will be sorted to the
  // beginning of the vector and the other to the end. If it will be
  // sort to the beginning, fix things up.
  if (lo_ieee_signbit (octave_NaN))
    {
      unsigned int i = 0;
      while (xisnan(x[i++]));
      for (unsigned int j = 0; j < elements - i + 1; j++)
	x[j] = x[j+i-1];
      for (unsigned int j = elements - i + 1; j < elements; j++)
	x[j] = octave_NaN;
    }
}

static void 
sort_with_index(double *x, double *tmp, int elements, int *indx, int *itmp)
{
  uint64 *sort = (uint64*)tmp;
  uint64 *array = (uint64*)x;

#define ARRAY_TO_TMP_EXTRA(pos) itmp[(pos)] = indx[i];
#define TMP_TO_ARRAY_EXTRA(pos) indx[(pos)] = itmp[i];

  DO_REAL_SORT (elements, array, sort);

#undef ARRAY_TO_TMP_EXTRA
#undef TMP_TO_ARRAY_EXTRA

  // There are two representations of NaN. One will be sorted to the
  // beginning of the vector and the other to the end. If it will be
  // sort to the beginning, fix things up.
  if (lo_ieee_signbit (octave_NaN))
    {
      unsigned int i = 0;
      while (xisnan(x[i]))
	  itmp[i] = indx[i++]; 
      for (unsigned int j = 0; j < elements - i; j++)
	{
	  x[j] = x[j+i];
	  indx[j] = indx[j+i];
	}
      for (unsigned int j = elements - i, k = 0; j < elements; j++, k++)
	{
	  x[j] = octave_NaN;
	  indx[j] = itmp[k++];
	}
    }
}

static octave_value_list
vec_sort (RowVector &v, bool return_idx)
{
  octave_value_list retval;
  int n = v.capacity ();

  if (n == 1)
    {
      if (return_idx)
	retval(1) = RowVector (n, 1.0);
      retval(0) = v;
      return retval;
    }
  else if (n > 1)
    {
      OCTAVE_LOCAL_BUFFER (double, tmp, n);

      if (return_idx)
	{
	  RowVector idx (n);
	  Array<int> l = create_index_array (n);

	  OCTAVE_LOCAL_BUFFER (int, itmp, n);

	  sort_with_index (v.fortran_vec (), tmp, n, 
			   l.fortran_vec (), itmp);
	  for (int i = 0; i < n; i++)
	    idx(i) = l(i);

	  retval(1) = idx;
	  retval(0) = v;
	}
      else
	{
	  sort (v.fortran_vec (), tmp, n);
	  retval(0) = v;
	}
    }

  return retval;
}

static octave_value_list
mx_sort (Matrix &m, bool return_idx)
{
  octave_value_list retval;
  int nr = m.rows ();
  int nc = m.columns ();

  if (nr == 1 && nc > 0)
    {
      if (return_idx)
	retval(1) = Matrix (nr, nc, 1.0);
      retval(0) = m;
      return retval;
    }
  else if (nr > 1 && nc > 0)
    {
      double * mv = m.fortran_vec ();
      OCTAVE_LOCAL_BUFFER (double, tmp, nr);

      if (return_idx)
	{
	  Matrix idx (nr, nc);

	  OCTAVE_LOCAL_BUFFER (int, itmp, nr);

	  for (int j = 0; j < nc; j++)
	    {
	      Array<int> l = create_index_array (nr);

	      sort_with_index (mv, tmp, nr, l.fortran_vec (), itmp);
	      for (int i = 0; i < nr; i++)
		idx(i,j) = l(i);
	      mv += nr;
	    }

	  retval(1) = idx;
	  retval(0) = m;
	}
      else
	{
	  for (int j = 0; j < nc; j++)
	    {
	      sort (mv, tmp, nr);
	      mv += nr;
	    }
	  retval(0) = m;
	}

    }

  return retval;
}


DEFUN_DLD (radix, args, nargout,
  "-*- texinfo -*-\n\
@deftypefn {Loadable Function} {[@var{s}, @var{i}] =} sort (@var{x})\n\
Return a copy of @var{x} with the elements elements arranged in\n\
increasing order.  For matrices, @code{sort} orders the elements in each\n\
column.\n\
\n\
For example,\n\
\n\
@example\n\
@group\n\
sort ([1, 2; 2, 3; 3, 1])\n\
     @result{}  1  1\n\
         2  2\n\
         3  3\n\
@end group\n\
@end example\n\
\n\
The @code{sort} function may also be used to produce a matrix\n\
containing the original row indices of the elements in the sorted\n\
matrix.  For example,\n\
\n\
@example\n\
@group\n\
[s, i] = sort ([1, 2; 2, 3; 3, 1])\n\
     @result{} s = 1  1\n\
            2  2\n\
            3  3\n\
     @result{} i = 1  3\n\
            2  1\n\
            3  2\n\
@end group\n\
@end example\n\
\n\
The for equal elements, the indices are such that the equal elements\n\
are listed in the order that appeared in the original list.\n\
\n\
@code{sort} can also be used to sort complex vectors and matrices. The\n\
values are sorted with respect to their magnitude. For equal magnitudes,\n\
the phase of the value determines the position of the returned element.\n\
If the phases are also equal, the position of the equal elements in the\n\
original list determines their position in the returned indices.\n\
@end deftypefn")
{
  octave_value_list retval;

  int nargin = args.length ();

  if (nargin != 1)
    {
      print_usage ("sort");
      return retval;
    }

  bool return_idx = nargout > 1;

  octave_value arg = args(0);

  if (arg.is_real_type ())
    {
      Matrix m = arg.matrix_value ();

      if (! error_state)
	{
	  if (m.rows () == 1)
	    {
	      int nc = m.columns ();
	      RowVector v (nc);
	      for (int i = 0; i < nc; i++)
		v(i) = m(0,i);

	      retval = vec_sort (v, return_idx);
	    }
	  else
	    retval = mx_sort (m, return_idx);
	}
    }
  else if (arg.is_complex_type ())
    error ("radix: can not sort complex type yet");
  else
    gripe_wrong_type_arg ("sort", arg);

  return retval;
}

/*
;;; Local Variables: ***
;;; mode: C++ ***
;;; End: ***
*/
